#lang racket/base

(require
 "1-lc-syntax.rkt"
 "2-anf-and-monadic-form.rkt"
 (except-in "3-machine-semantics.rkt" insert deref)
 (except-in "4-ab-normal-form.rkt" insert deref)
 "5-compiler.rkt"
 redex/reduction-semantics
 redex/gui)

(provide
 apply-reduction-relation
 apply-reduction-relation*
 redex-match?
 redex-match
 term
 traces
 hole

 lcL
 eval-lcL
 (rename-out
  [lcL lambdaL]
  [eval-lcL eval-lambdaL])
 δ

 intro-example
 intro-example-monadic
 intro-example-anf
 running-example
 nested-branches
 (rename-out
  [nested-branches nested-branches-example])
 fact-example
 tail-fact-example

 lCKL
 (rename-out
  [lCKL lambda-ckL])
 lck->
 (rename-out
  [lck-> lambda-ck->])
 init-ck

 a->
 a->*
 a-normalize
 ANFL
 meval-lcL
 (rename-out
  [meval-lcL monadic-eval-lambdaL])
 b->
 b->*
 b-normalize
 monadicL

 anfck->
 bnfck->
 (rename-out
  [anfck-> anf-ck->]
  [bnfck-> bnf-ck->])
 ck-max-stack

 regionL
 regionCSKL
 regioncsk->
 (rename-out
  [regioncsk-> region-csk->])
 init-csk
 csk-read-val
 csk-max-memory
 csk-max-regions

 reg1-term

 aregionL
 ra->
 ra->*
 anf-reg1-term

 bregionL
 rb->
 rb->*
 bnf-reg1-term

 imonadicL
 abnfL
 ab->
 ab->*
 ab-normalize
 cek-imonadicL
 (rename-out
  [cek-imonadicL imonadic-cekL])
 init-cek
 cek->
 cek-max-stack
 monadic-large-eg-term

 rimonadicL
 cesk->
 init-cesk
 rab->
 rab->*
 rab-normalize
 flatten-begin
 cesk-max-memory
 cesk-max-stack
 cesk-max-regions
 cesk-read-val
 lst:code-gen-mregion-eg1

 anf
 monadic
 monadic-code-gen
 anf-code-gen
 abnf

 abnormal-compile
 anf-compile
 inc-var

 test-theorem-3.1?
 test-theorem-3.2?
 test-theorem-3.3?
 test-theorem-4.1?
 test-theorem-4.2?
 )

(require racket/contract)

(define/contract (test-theorem-3.1? e)
 (-> (redex-match? monadicL C) #t)
 (and
  (>=
   (ck-max-stack (apply-reduction-relation* #:cache-all? #t bnfck-> (init-ck e)))
   (ck-max-stack (apply-reduction-relation* #:cache-all? #t anfck-> (init-ck (a-normalize e)))))
  ; witness for strictness in some case
  (>
   (ck-max-stack (apply-reduction-relation* #:cache-all? #t bnfck-> (init-ck (b-normalize running-example))))
   (ck-max-stack (apply-reduction-relation* #:cache-all? #t anfck-> (init-ck (a-normalize running-example)))))))

(define/contract (test-theorem-3.2? C)
 (-> (redex-match? bregionL C) #t)
 (and
  (>=
   (csk-max-memory (apply-reduction-relation* #:all? #t regioncsk-> (init-csk (car (apply-reduction-relation* #:cache-all? #t ra->* C)))))
   (csk-max-memory (apply-reduction-relation* #:all? #t regioncsk-> (init-csk C))))

  ; witness for strictness in some case
  (>
   (csk-max-memory (apply-reduction-relation* #:all? #t regioncsk-> (init-csk (car (apply-reduction-relation* #:cache-all? #t ra->* reg1-term)))))
   (csk-max-memory (apply-reduction-relation* #:all? #t regioncsk-> (init-csk reg1-term))))

  (>=
   (csk-max-regions (apply-reduction-relation* #:all? #t regioncsk-> (init-csk (car (apply-reduction-relation* #:cache-all? #t ra->* C)))))
   (csk-max-regions (apply-reduction-relation* #:all? #t regioncsk-> (init-csk C))))

  ; witness for strictness in some case
  (>
   (csk-max-regions (apply-reduction-relation* #:all? #t regioncsk-> (init-csk (car (apply-reduction-relation* #:cache-all? #t ra->* reg1-term)))))
   (csk-max-regions (apply-reduction-relation* #:all? #t regioncsk-> (init-csk reg1-term))))))

(define/contract (test-theorem-3.3? e)
 (-> (redex-match? regionL e) #t)
 (and
  (=
   (csk-max-memory (apply-reduction-relation* #:all? #t regioncsk-> (init-csk (car (apply-reduction-relation* #:cache-all? #t rb->* e)))))
   (csk-max-memory (apply-reduction-relation* #:all? #t regioncsk-> (init-csk e))))
  (=
   (csk-max-regions (apply-reduction-relation* #:all? #t regioncsk-> (init-csk (car (apply-reduction-relation* #:cache-all? #t rb->* e)))))
   (csk-max-regions (apply-reduction-relation* #:all? #t regioncsk-> (init-csk e))))))

(define/contract (test-theorem-4.1? t)
 (-> (redex-match? imonadicL t) #t)
 (and
  (>=
   (cek-max-stack (apply-reduction-relation* #:all? #t cek-> (init-cek t)))
   (cek-max-stack (apply-reduction-relation* #:all? #t cek-> (init-cek (ab-normalize t)))))

  ; witness for strictness in some case
  (>
   (cek-max-stack (apply-reduction-relation* #:all? #t cek-> (init-cek monadic-large-eg-term)))
   (cek-max-stack (apply-reduction-relation* #:all? #t cek-> (init-cek (ab-normalize monadic-large-eg-term)))))))

(define/contract (test-theorem-4.2? t)
 (-> (redex-match? rimonadicL t) #t)
 (and
  (>=
   (cesk-max-stack (apply-reduction-relation* #:all? #t cesk-> (init-cesk t)))
   (cesk-max-stack (apply-reduction-relation* #:all? #t cesk-> (init-cesk (abnf t)))))

  ; witness for strictness in some case
  (>
   (cesk-max-stack (apply-reduction-relation* #:all? #t cesk-> (init-cesk lst:code-gen-mregion-eg1)))
   (cesk-max-stack (apply-reduction-relation* #:all? #t cesk-> (init-cesk (abnf lst:code-gen-mregion-eg1)))))

  (=
   (cesk-max-memory (apply-reduction-relation* #:all? #t cesk-> (init-cesk t)))
   (cesk-max-memory (apply-reduction-relation* #:all? #t cesk-> (init-cesk (abnf t)))))

  (=
   (cesk-max-regions (apply-reduction-relation* #:all? #t cesk-> (init-cesk t)))
   (cesk-max-regions (apply-reduction-relation* #:all? #t cesk-> (init-cesk (abnf t)))))))
